/*
 * Copyright 2015-present Facebook, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once

#include <memory>
#include <chrono>

#include <thrift/lib/cpp2/server/ThriftServer.h>
#include <thrift/lib/cpp2/test/util/TestServerFactory.h>

namespace apache { namespace thrift {

template<typename Interface>
struct TestThriftServerFactory : public TestServerFactory {
  public:
   std::shared_ptr<BaseThriftServer> create() override {
     auto server = std::make_shared<apache::thrift::ThriftServer>();
     server->setNumIOWorkerThreads(1);
     if (useSimpleThreadManager_) {
       auto threadFactory =
           std::make_shared<apache::thrift::concurrency::PosixThreadFactory>();
       auto threadManager =
           apache::thrift::concurrency::ThreadManager::newSimpleThreadManager(
               1, false);
       threadManager->threadFactory(threadFactory);
       threadManager->start();
       server->setThreadManager(threadManager);
     } else if (exe_) {
       server->setThreadManager(exe_);
     }

     server->setPort(0);

     if (idleTimeoutMs_ != 0) {
       server->setIdleTimeout(std::chrono::milliseconds(idleTimeoutMs_));
     }

     if (duplex_) {
       server->setDuplex(true);
     }

     if (serverEventHandler_) {
       server->setServerEventHandler(serverEventHandler_);
     }

     server->setMinCompressBytes(minCompressBytes_);
     if (transId_) {
       std::vector<uint16_t> transforms{transId_};
       server->setDefaultWriteTransforms(transforms);
     }
     server->setInterface(std::unique_ptr<Interface>(new Interface));
     return server;
  }

  TestThriftServerFactory& useSimpleThreadManager(bool use) override {
    useSimpleThreadManager_ = use;
    return *this;
  }

  TestThriftServerFactory& useThreadManager(
      std::shared_ptr<apache::thrift::concurrency::ThreadManager> exe)
      override {
    exe_ = exe;
    return *this;
  }

  TestThriftServerFactory& idleTimeoutMs (uint32_t idle) {
    idleTimeoutMs_ = idle;
    return *this;
  }

  TestThriftServerFactory& duplex(bool duplex) {
    duplex_ = duplex;
    return *this;
  }

  TestThriftServerFactory& minCompressBytes(uint32_t minCompressBytes) {
    minCompressBytes_ = minCompressBytes;
    return *this;
  }

  TestThriftServerFactory& defaultWriteTransform(uint16_t transId) {
    transId_ = transId;
    return *this;
  }

 private:
  bool useSimpleThreadManager_{true};
  std::shared_ptr<apache::thrift::concurrency::ThreadManager> exe_{nullptr};
  uint32_t idleTimeoutMs_{0};
  bool duplex_{false};
  uint32_t minCompressBytes_{0};
  uint16_t transId_{0};
};

}} // apache::thrift
